/*
 * Copyright (c) 1995  Shigeki Miura
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice in the documentation and/or other materials provided with 
 *    the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN 
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*
 * Routines which deal with the characteristics of the terminal.
 *
 * This file is specific to Windows NT Console.
 */

#include "less.h"
#include "cmd.h"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

static int term_init = 0;
static int init_done = 0;

public int auto_wrap;		/* Terminal does \r\n when write past margin */
public int ignaw;		/* Terminal ignores \n immediately after wrap */
public int can_goto_line;
public int erase_char, kill_char; /* The user's erase and line-kill chars */
public int sc_width, sc_height;	  /* Height & width of screen */
public int bo_s_width, bo_e_width; /* Printing width of boldface seq */
public int ul_s_width, ul_e_width; /* Printing width of underline seq */
public int so_s_width, so_e_width; /* Printing width of standout seq */
public int bl_s_width, bl_e_width; /* Printing width of blink seq */

public HANDLE hConsole;

public WORD nm_color;		/* Color of normal text */
public WORD bo_color;		/* Color of bold text */
public WORD ul_color;		/* Color of underline text */
public WORD so_color;		/* Color of standout text */
public WORD bl_color;		/* Color of blinking text */

static WORD sy_color;
static WORD *pAttributes = NULL;

extern int quiet;		/* If VERY_QUIET, use visual bell for bell */

/*
 * Change terminal to "raw mode", or restore to "normal" mode.
 * "Raw mode" means 
 *	1. An outstanding read will complete on receipt of a single keystroke.
 *	2. Input is not echoed.  
 *	3. On output, \n is mapped to \r\n.
 *	4. \t is NOT expanded into spaces.
 *	5. Signal-causing characters such as ctrl-C (interrupt),
 *	   etc. are NOT disabled.
 * It doesn't matter whether an input \n is mapped to \r, or vice versa.
 */
	public void
raw_mode (on)
     int on;
{
  static int curr_on = 0;
  extern HANDLE hConsoleIn;

  if (on == curr_on)
    return;
  erase_char = CONTROL('h');	/* ^H */
  kill_char = CONTROL('u');	/* ^U */

  if (on)
    SetConsoleMode (hConsoleIn, ENABLE_PROCESSED_INPUT);
  else
    SetConsoleMode (hConsoleIn,
		    (ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT
		     | ENABLE_PROCESSED_INPUT | ENABLE_MOUSE_INPUT));
  curr_on = on;
}

/*
 * Get size of the output screen.
 */
	public void
scrsize (void)
{
  CONSOLE_SCREEN_BUFFER_INFO info;

  if (GetConsoleScreenBufferInfo (hConsole, &info))
    {
      sc_width = info.dwSize.X;
      sc_height = info.dwSize.Y;
    }
  else
    {
      sc_width = 80;
      sc_height = 25;
    }
}

/*
 * Table of line editting characters, for editchar() in decode.c.
 */
char edittable[] = {
  '\340',VK_RIGHT,0,		EC_RIGHT,	/* RIGHTARROW */
  '\340',VK_LEFT,0,		EC_LEFT,	/* LEFTARROW */
  '\340',VK_CONTROL,VK_LEFT,0,	EC_W_LEFT,	/* CTRL-LEFTARROW */
  '\340',VK_CONTROL,VK_RIGHT,0,	EC_W_RIGHT,	/* CTRL-RIGHTARROW */
  '\340',VK_INSERT,0,		EC_INSERT,	/* INSERT */
  '\340',VK_DELETE,0,		EC_DELETE,	/* DELETE */
  '\340',VK_CONTROL,VK_DELETE,0,EC_W_DELETE,	/* CTRL-DELETE */
  '\177',0,			EC_W_BACKSPACE,	/* CTRL-BACKSPACE */
  '\340',VK_HOME,0,		EC_HOME,	/* HOME */
  '\340',VK_END,0,		EC_END,		/* END */
  '\340',VK_UP,0,		EC_UP,		/* UPARROW */
  '\340',VK_DOWN,0,		EC_DOWN,	/* DOWNARROW */
  '\t',0,			EC_F_COMPLETE,	/* TAB */
  '\17',0,			EC_B_COMPLETE,	/* BACKTAB (?) */
  '\340','\17',0,		EC_B_COMPLETE,	/* BACKTAB */
  '\14',0,			EC_EXPAND,	/* CTRL-L */
};

char cmdtable[] =
{
  /*
   * PC function keys.
   * Note that '\0' is converted to '\340' on input.
   */
  '\340',VK_DOWN,0,	A_F_LINE,	/* down arrow */
  '\340',VK_NEXT,0,	A_F_SCREEN,	/* page down */
  '\340',VK_UP,0,	A_B_LINE,	/* up arrow */
  '\340',VK_PRIOR,0,	A_B_SCREEN,	/* page up */
  '\340',VK_HOME,0,	A_GOLINE,	/* home */
  '\340',VK_END,0,	A_GOEND,	/* end */
  '\340',VK_F1,0,	A_HELP,		/* F1 */
};

	public void
get_editkeys (void)
{
  add_fcmd_table (cmdtable, sizeof (cmdtable));
  add_ecmd_table (edittable, sizeof (edittable));
}

/*
 * Get terminal capabilities via termcap.
 */
	public void
get_term (void)
{
  CONSOLE_SCREEN_BUFFER_INFO info;

  if (!term_init)
    {
      term_init = 1;
      hConsole = GetStdHandle (STD_OUTPUT_HANDLE);

      GetConsoleScreenBufferInfo (hConsole, &info);
      nm_color = info.wAttributes;
      bo_color = info.wAttributes | FOREGROUND_INTENSITY;
      ul_color = info.wAttributes | COMMON_LVB_UNDERSCORE;
      so_color = info.wAttributes | COMMON_LVB_REVERSE_VIDEO;
      bl_color = (info.wAttributes
		  | BACKGROUND_BLUE | BACKGROUND_GREEN | BACKGROUND_RED);
    }

  scrsize ();
  pos_init ();
  auto_wrap = 1;
  ignaw = 0;
  can_goto_line = 1;
  so_e_width = so_s_width = 0;
  bo_s_width = bo_e_width = 0;
  ul_s_width = ul_e_width = 0;
  bl_s_width = bl_e_width = 0;
}


/*
 * Below are the functions which perform all the 
 * terminal-specific screen manipulation.
 */


/*
 * Initialize terminal
 */
	public void
init (void)
{
  CONSOLE_SCREEN_BUFFER_INFO info;

  /* {{ What could we take no_init (-X) to mean? }} */
  GetConsoleScreenBufferInfo (hConsole, &info);
  sy_color = info.wAttributes;
  SetConsoleTextAttribute (hConsole, nm_color);

  pAttributes = malloc (sizeof (WORD) * sc_width * sc_height);

  flush ();
  init_done = 1;
}

/*
 * Deinitialize terminal
 */
	public void
deinit (void)
{
  if (!init_done)
    return;
  putstr ("\n");
  if (pAttributes)
    free (pAttributes);
  SetConsoleTextAttribute (hConsole, sy_color);
  init_done = 0;
}

/*
 * Home cursor (move to upper left corner of screen).
 */
	public void
home (void)
{
  COORD pos;

  flush ();
  pos.X = pos.Y = 0;
  SetConsoleCursorPosition (hConsole, pos);
}

/*
 * Add a blank line (called with cursor at home).
 * Should scroll the display down.
 */
	public void
add_line (void)
{
  COORD pos, dest;
  SMALL_RECT rect;
  CHAR_INFO fill;

  rect.Left = 0;
  rect.Top = 0;
  rect.Right = sc_width;
  rect.Bottom = sc_height;
  fill.Char.AsciiChar = ' ';
  fill.Attributes = nm_color;
  dest.X = 0;
  dest.Y = 1;
  pos.X = pos.Y = 0;

  flush ();
  ScrollConsoleScreenBuffer (hConsole, &rect, NULL, dest, &fill);
  SetConsoleCursorPosition (hConsole, pos);
}

/*
 * Move cursor to lower left corner of screen.
 */
	public void
lower_left (void)
{
  COORD pos;

  flush ();
  pos.X = 0;
  pos.Y = sc_height - 1;
  SetConsoleCursorPosition (hConsole, pos);
}

/*
 * Goto a specific line on the screen.
 */
	public void
goto_line (slinenum)
     int slinenum;
{
  COORD pos;

  flush ();
  pos.X = 0;
  pos.Y = slinenum;
  SetConsoleCursorPosition (hConsole, pos);
}

/*
 * Output the "visual bell", if there is one.
 */
	public void
vbell (void)
{
  DWORD dwRead, dwWritten;
  COORD pos;

  if (pAttributes)
    {
      pos.X = pos.Y = 0;
      ReadConsoleOutputAttribute (hConsole, pAttributes, sc_width * sc_height,
				  pos, &dwRead);
      FillConsoleOutputAttribute (hConsole, so_color, sc_width * sc_height,
				  pos, &dwWritten);
      Sleep (100);
      WriteConsoleOutputAttribute (hConsole, pAttributes, sc_width * sc_height,
				   pos, &dwWritten);
    }
}

/*
 * Ring the terminal bell.
 */
	public void
bell (void)
{
  if (quiet == VERY_QUIET)
    vbell ();
  else
    Beep (666, 100);
}

/*
 * Clear the screen.
 */
	public void
clear (void)
{
  CONSOLE_SCREEN_BUFFER_INFO info;
  COORD pos;
  DWORD dwWritten;

  flush ();
  GetConsoleScreenBufferInfo (hConsole, &info);
  pos.X = pos.Y = 0;
  FillConsoleOutputAttribute (hConsole, nm_color,
			      (info.dwMaximumWindowSize.X
			       * info.dwMaximumWindowSize.Y),
			      pos, &dwWritten);
  FillConsoleOutputCharacter (hConsole, ' ',
			      (info.dwMaximumWindowSize.X
			       * info.dwMaximumWindowSize.Y),
			      pos, &dwWritten);
}

/*
 * Clear from the cursor to the end of the cursor's line.
 * {{ This must not move the cursor. }}
 */
	public void
clear_eol (void)
{
  CONSOLE_SCREEN_BUFFER_INFO info;
  DWORD dwWritten;

  flush ();
  GetConsoleScreenBufferInfo (hConsole, &info);
  FillConsoleOutputAttribute (hConsole, nm_color,
			      sc_width - info.dwCursorPosition.X,
			      info.dwCursorPosition,
			      &dwWritten);
  FillConsoleOutputCharacter (hConsole, ' ',
			      sc_width - info.dwCursorPosition.X,
			      info.dwCursorPosition,
			      &dwWritten);
}

/*
 * Clear the bottom line of the display.
 * Leave the cursor at the beginning of the bottom line.
 */
	public void
clear_bot (void)
{
  lower_left ();
  clear_eol ();
}

/*
 * Begin "standout" (bold, underline, or whatever).
 */
	public void
so_enter (void)
{
  flush ();
  SetConsoleTextAttribute (hConsole, so_color);
}

/*
 * End "standout".
 */
	public void
so_exit (void)
{
  flush ();
  SetConsoleTextAttribute (hConsole, nm_color);
}

/*
 * Begin "underline" (hopefully real underlining, 
 * otherwise whatever the terminal provides).
 */
	public void
ul_enter (void)
{
  flush ();
  SetConsoleTextAttribute (hConsole, ul_color);
}

/*
 * End "underline".
 */
	public void
ul_exit (void)
{
  flush ();
  SetConsoleTextAttribute (hConsole, nm_color);
}

/*
 * Begin "bold"
 */
	public void
bo_enter (void)
{
  flush ();
  SetConsoleTextAttribute (hConsole, bo_color);
}

/*
 * End "bold".
 */
	public void
bo_exit (void)
{
  flush ();
  SetConsoleTextAttribute (hConsole, nm_color);
}

/*
 * Begin "blink"
 */
	public void
bl_enter (void)
{
  flush ();
  SetConsoleTextAttribute (hConsole, bl_color);
}

/*
 * End "blink".
 */
	public void
bl_exit (void)
{
  flush ();
  SetConsoleTextAttribute (hConsole, nm_color);
}

/*
 * Erase the character to the left of the cursor 
 * and move the cursor left.
 */
	public void
backspace (void)
{
  CONSOLE_SCREEN_BUFFER_INFO info;
  DWORD dwWritten;

  flush ();
  GetConsoleScreenBufferInfo (hConsole, &info);
  if (info.dwCursorPosition.X <= 0)
    return;

  info.dwCursorPosition.X--;
  FillConsoleOutputAttribute (hConsole, nm_color, 1,
			      info.dwCursorPosition, &dwWritten);
  FillConsoleOutputCharacter (hConsole, ' ', 1,
			      info.dwCursorPosition, &dwWritten);
  SetConsoleCursorPosition (hConsole, info.dwCursorPosition);
}

/*
 * Output a plain backspace, without erasing the previous char.
 */
	public void
putbs (void)
{
  CONSOLE_SCREEN_BUFFER_INFO info;

  flush ();

  GetConsoleScreenBufferInfo (hConsole, &info);

  if (info.dwCursorPosition.X <= 0)
    return;

  info.dwCursorPosition.X--;
  SetConsoleCursorPosition (hConsole, info.dwCursorPosition);
}

	public void
move_right (int n)
{
  CONSOLE_SCREEN_BUFFER_INFO info;

  flush ();

  GetConsoleScreenBufferInfo (hConsole, &info);
  info.dwCursorPosition.X += n;
  SetConsoleCursorPosition (hConsole, info.dwCursorPosition);
}
